import json
import os
import re

import requests


def find_file(name, path):
    """
    Finds file with the specified name in the given path.

    Args:
        name (str): file name.
        path (str): starting path for search.

    Returns:
        str: absolute file path (if found)
    """
    for root, dirs, files in os.walk(path, followlinks=True):
        for file in files:
            if name in file:
                return os.path.join(root, file)


def find_file_with_pattern(pattern, path):
    """
    Finds a file inside a given directory that matches the pattern.

    Args:
        path (str): starting path for search.
        pattern (str): regex to apply on the file content.

    Returns:
        str: absolute file path
    """
    for root, dirs, files in os.walk(path, followlinks=True):
        for file_ in files:
            with open(os.path.join(root, file_)) as f:
                if re.match(pattern, f.read()):
                    return os.path.join(root, file_)


def read(path):
    """
    Reads and returns the content of given file.

    Args:
        path (str): file path.

    Returns:
        str: file content.
    """
    with open(path) as f:
        return f.read()


def write_json(path, content):
    """
    Writes a given JSON content.

    Args:
        path (str): file path.
        content (dict): content.
    """
    with open(path, "w+") as f:
        f.write(json.dumps(content, indent=4))


def read_json(path):
    """
    Reads and returns the content of given file in JSON format.

    Args:
        path (str): file path.

    Returns:
        dict: file content.
    """
    return json.loads(read(path))


def upload_file_to_object_storage(file_):
    """
    Uploads a given file to object storage.

    Note: a new line is added to the file if it is empty.

    Args:
        file_ (dict): a dictionary with path and presignedURL keys.
    """
    path = file_["path"]
    session = requests.Session()
    if os.path.getsize(path) == 0:
        with open(path, 'w+') as f:
            f.write('\n')
    headers = {"Content-Length": str(os.path.getsize(path))}
    with open(path) as f:
        session.request("PUT", file_["URL"], data=f, headers=headers).raise_for_status()



ESPRESSO_EXECUTABLE_NAME_REGEX = r'\s+Program (.*) v.* starts on .*'
ESPRESSO_INPUT_FILE_REGEX = r'&CONTROL|&SYSTEM|&ELECTRONS|&IONS|&CELL|&BANDS|&INPUT|&PROJWFC|&DOS|'

ESPRESSO_EXECUTABLE_NAME_MAP = {
    "PWSCF": "pw.x",
    "BANDS": "bands.x",
    "MATDYN": "matdyn.x",
    "DYNMAT": "dynmat.x",
    "Q2R": "q2r.x",
    "PHONON": "ph.x",
    "PROJWFC": "projwfc.x",
    "DOS": "dos.x",
    "NEB": "neb.x",
}

ESPRESSO_DEFAULT_VERSION = "5.4.0"
ESPRESSO_SUPPORTED_VERSIONS = ["5.2.1", "5.4.0", "6.0.0", "6.3"]

VASP_DEFAULT_VERSION = "5.4.4"
VASP_SUPPORTED_VERSIONS = ["5.3.5", "5.4.4"]

SHELL_DEFAULT_VERSION = "4.2.46"
SHELL_SUPPORTED_VERSIONS = ["4.2.46"]

OUTPUT_CHUNK_SIZE = 1024 * 1024  # in bytes



import os

from exaparser.config import ExaParserConfig
from exaparser.utils import write_json
from .. import DataHandler


class DiskDataHandler(DataHandler):
    """
    Disk data handler class.

    Args:
        job (exaparser.job.Job)
    """

    def __init__(self, job):
        super(DiskDataHandler, self).__init__(job)

    def handle(self):
        """
        Stores job inside data_dir directory.
        """
        data_dir = os.path.join(self.job.work_dir, ExaParserConfig.get("disk_data_handler", "data_dir"))
        if not os.path.exists(data_dir):
            os.makedirs(data_dir)
        write_json(os.path.join(data_dir, "job.json"), self.job.to_json())



import xml.etree.ElementTree as ET

from express.parsers.apps.vasp.settings import XML_DATA_FILE as VASP_XML_FILE

from exaparser.enums import *
from exaparser.utils import find_file
from . import ModelingExecutionUnit


class VaspExecutionUnit(ModelingExecutionUnit):
    """
    Vasp execution unit parser class.

    Args:
        config (dict): unit config.
        work_dir (str): full path to working directory.
    """

    def __init__(self, config, work_dir):
        super(VaspExecutionUnit, self).__init__(config, work_dir)
        self.xml_path = find_file(VASP_XML_FILE, self.work_dir)

    @property
    def parser_name(self):
        """
        Returns the name of the parser to pass to ExPrESS.

        Returns:
             str
        """
        return "vasp"

    @property
    def application(self):
        """
        Returns the application used in the unit.

        Returns:
             dict
        """
        return {
            "name": "vasp",
            "version": self.version,
            "summary": "Vienna Ab-initio Simulation Package"
        }

    @property
    def version(self):
        """
        Returns the application version used in the unit.

        Returns:
             str
        """
        root = ET.parse(self.xml_path).getroot()
        version = root.find('generator').find('.//i[@name="version"]').text.strip()
        return version if version in VASP_SUPPORTED_VERSIONS else VASP_DEFAULT_VERSION

    @property
    def executable(self):
        """
        Returns the executable used in the unit.

        Returns:
             dict
        """
        return {
            "name": "vasp"
        }



from .units.factory import get_unit


class Subworkflow(object):
    """
    Subworkflow parser class.

    Args:
        config (dict): subworkflow config.
        work_dir (str): full path to working directory.
    """

    def __init__(self, config, work_dir):
        self.config = config
        self.work_dir = work_dir
        self.units = [get_unit(config, self.work_dir) for config in self.config.get("units", [])]

    @property
    def id(self):
        """
        Returns subworkflow id.

        Returns:
             str
        """
        return self.config["_id"]

    @property
    def name(self):
        """
        Returns subworkflow name.

        Returns:
             str
        """
        return self.config.get("name", "Subworkflow")

    @property
    def application(self):
        """
        Returns the application used in the subworkflow.

        Note: This is a placeholder for future to extract application from disk.

        Returns:
             dict
        """
        return self.units[0].application

    @property
    def model(self):
        """
        Returns the model used in the subworkflow.

        Note: This is a placeholder for future to extract model from disk.

        Returns:
             dict
        """
        return self.config["model"]

    @property
    def execution_units(self):
        """
        Returns a list of all execution units in this subworkflow.

        Returns:
             list
        """
        execution_units = []
        for unit in self.units:
            if unit.type == "execution":
                execution_units.append(unit)
        return execution_units

    def to_json(self):
        """
        Returns the subworkflow in JSON format.

        Returns:
             dict
        """
        return {
            "_id": self.id,
            "name": self.name,
            "model": self.model,
            "properties": [],
            "application": self.application,
            "units": [u.to_json() for u in self.units]
        }



import os

from exaparser.config import ExaParserConfig
from exaparser.utils import read_json
from .compute.factory import get_compute_parser
from ..workflow.workflow import Workflow


class Job(object):
    """
    Job parser class.

    Args:
        work_dir (str): full path to working directory.
    """

    def __init__(self, name, work_dir):
        self.name = name
        self.work_dir = work_dir
        self._workflow = None
        self._compute = None

    @property
    def compute(self):
        """
        Returns compute parser class to extract compute configuration.
        """
        if not self._compute:
            self._compute = get_compute_parser(ExaParserConfig["global"]["rms_type"], self.work_dir)
        return self._compute

    @property
    def structures(self):
        """
        Returns a list of structures (initial/final) used in job.

        Returns:
             list[dict]
        """
        structures = []
        for unit in self.workflow.execution_units:
            structures.extend(getattr(unit, "structures", []))
        return structures

    @property
    def stdout_files(self):
        """
        Returns a list of stdout files for all execution units.

        Returns:
             list[dict]
        """
        stdout_files = []
        for unit in self.workflow.execution_units:
            if os.path.exists(unit.stdout_file):
                stdout_files.append({
                    "stdoutFile": unit.stdout_file,
                    "unitFlowchartId": unit.flowchartId,
                })
        return stdout_files

    @property
    def status(self):
        """
        Returns job status.
        Status is set to "error" if there is a unit in "error" status.

        Returns:
             str
        """
        status = "finished"
        for unit in self.workflow.execution_units:
            if unit.status == "error":
                status = "error"
        return status

    @property
    def properties(self):
        """
        Returns a list of all properties extracted from the job.

        Returns:
             list[dict]
        """
        properties = []
        for unit in self.workflow.execution_units:
            properties.extend(getattr(unit, "properties", []))
        return properties

    @property
    def workflow(self):
        """
        Returns an instance of Workflow class.
        """
        if not self._workflow:
            templates_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), "templates")
            template_path = os.path.join(templates_dir, ExaParserConfig["global"]["workflow_template_name"])
            self._workflow = Workflow(read_json(template_path), self.work_dir)
        return self._workflow

    def to_json(self):
        """
        Returns the job in JSON format.

        Returns:
             dict
        """
        return {
            "_project": {
                "slug": ExaParserConfig["global"]["project_slug"]
            },
            "compute": self.compute.to_json(),
            "owner": {
                "slug": ExaParserConfig["global"]["owner_slug"]
            },
            "name": self.name,
            "status": self.status,
            "workDir": self.work_dir,
            "workflow": self.workflow.to_json()
        }



class DataHandler(object):
    """
    Base data handler class.

    Args:
        job (exaparser.job.Job): an instance of the job class.
    """

    def __init__(self, job):
        self.job = job

    def handle(self):
        """
        Implement the logic to handle extracted job, materials and properties.
        Override upon inheritance.
        """
        raise NotImplementedError



from .handlers.disk import DiskDataHandler
from .handlers.exabyte import ExabyteRESTFulAPIDataHandler
from .handlers.stdout import StdoutDataHandler


def get_data_handler(name, job):
    """
    Returns an instance of the data handler class.

    Args:
        name (str): data handler name, STDOUT, DISK or ExabyteRESTFulAPI
        job (exaparser.job.Job): an instance of the job class.
    """
    handlers = dict(STDOUT=StdoutDataHandler, DISK=DiskDataHandler, ExabyteRESTFulAPI=ExabyteRESTFulAPIDataHandler)
    return handlers[name](job)



import os

from exaparser.utils import read
from .. import BaseUnit


class BaseExecutionUnit(BaseUnit):
    """
    Base execution unit parser class.

    Args:
        config (dict): unit config.
        work_dir (str): full path to working directory.
    """

    def __init__(self, config, work_dir):
        super(BaseExecutionUnit, self).__init__(config, work_dir)
        self.work_dir = os.path.join(self.work_dir, self.config.get("workDir", ""))

    @property
    def stdout_file(self):
        return os.path.join(self.work_dir, self.config.get("stdoutFile", '.'.join((self.name, 'out'))))

    @property
    def application(self):
        """
        Returns the application used in the unit.
        Override upon inheritance.

        Returns:
             dict
        """
        raise NotImplementedError

    @property
    def version(self):
        """
        Returns the application version used in the unit.

        Returns:
             str
        """
        raise NotImplementedError

    @property
    def executable(self):
        """
        Returns the executable used in the unit.
        Override upon inheritance.

        Returns:
             dict
        """
        raise NotImplementedError

    @property
    def input(self):
        """
        Returns a list of input files used in the unit.

        Note: Make sure to set "isManuallyChanged" to True.

        Returns:
             list[dict]
        """
        input_ = []
        for config in self.config.get("input", []):
            path_ = os.path.join(self.work_dir, config["name"])
            input_.append({
                "name": config["name"],
                "isManuallyChanged": True,
                "rendered": read(path_) if os.path.exists(path_) else ""
            })
        return input_

    @property
    def postProcessors(self):
        """
        Returns a list of postProcessors used in the unit.
        Override upon inheritance as necessary.

        Returns:
             list[dict]
        """
        return self.config.get("postProcessors", [])

    @property
    def preProcessors(self):
        """
        Returns a list of preProcessors used in the unit.
        Override upon inheritance as necessary.

        Returns:
             list[dict]
        """
        return self.config.get("preProcessors", [])

    def to_json(self):
        """
        Returns the unit in JSON format.

        Returns:
             dict
        """
        config = super(BaseExecutionUnit, self).to_json()
        config.update({
            "application": self.application,
            "executable": self.executable,
            "input": self.input,
            "monitors": self.monitors,
            "name": self.name,
            "postProcessors": self.postProcessors,
            "preProcessors": self.preProcessors,
            "results": self.results,
            "type": "execution",
            "status": self.status,
            "statusTrack": self.status_track,
        })
        return config

    @property
    def status(self):
        """
        Returns unit status.

        Note: This is a placeholder for future to extract unit status based on unit outputs.

        Returns:
             str
        """
        return "finished"

    @property
    def status_track(self):
        """
        Returns unit status track.

        Returns:
             list[dict]
        """
        return [
            {
                "trackedAt": int(os.path.getmtime(self.work_dir)),
                "status": self.status
            }
        ]

    @property
    def results(self):
        """
        Returns a list of property names extracted from the unit.

        Returns:
             list[dict]
        """
        return self.config.get("results", [])

    @property
    def monitors(self):
        """
        Returns a list of monitors used in the unit.

        Returns:
             list[dict]
        """
        return self.config.get("monitors", [])



from .. import BaseComputeParser


class SLURMComputeParser(BaseComputeParser):
    """
    Slurm compute parser class.

    Args:
        work_dir (str): full path to working directory.
    """

    def __init__(self, work_dir):
        super(SLURMComputeParser, self).__init__(work_dir)



import json

from .. import DataHandler


class StdoutDataHandler(DataHandler):
    """
    Stdout data handler class.

    Args:
        job (exaparser.job.Job)
    """

    def __init__(self, job):
        super(StdoutDataHandler, self).__init__(job)

    def print_json(self, content):
        print(json.dumps(content, indent=4))

    def handle(self):
        """
        Prints the job in standard output.
        """
        self.print_json(self.job.to_json())



import configparser
import os
from pathlib import Path


def find_config_file() -> Path:
    """
    Look in an ordered list of standard locations for the exaparser config file, and return the first one we find.
    If no config file is found, throw an Exception()
    """
    possible_filepaths = [
        Path(__file__) / '../../config',
        Path(os.path.expanduser('~')) / '.exabyte/exaparser/config',
        Path('/etc/exabyte/exaparser/config'),
    ]
    for path in possible_filepaths:
        path = path.resolve()
        if path.is_file():
            return path
    raise Exception('Could not find exaparser config file in any of the known locations.')


ExaParserConfig = configparser.ConfigParser(allow_no_value=True)
ExaParserConfig.read(find_config_file())



from .execution.modeling.espresso import EspressoExecutionUnit
from .execution.modeling.vasp import VaspExecutionUnit
from .execution.scripting.shell import ShellExecutionUnit
from .execution.scripting.shell_with_results import ShellWithResultsExecutionUnit
from .subworkflow import SubworkflowUnit


def get_unit(config, work_dir):
    """
    Returns a unit by given config.

    Args:
        config (dict): unit configuration obtained from the template.
        work_dir (str): full path to the job working directory.
    """
    if config["type"] == "execution":
        execution_units = dict(shell=ShellExecutionUnit, vasp=VaspExecutionUnit, espresso=EspressoExecutionUnit)
        # Use ShellWithResultsExecutionUnit if parser is asked to extract any results
        if config["application"]["name"] == "shell" and len(config.get("results", [])):
            return ShellWithResultsExecutionUnit(config, work_dir)
        return execution_units[config["application"]["name"]](config, work_dir)

    if config["type"] == "subworkflow":
        return SubworkflowUnit(config, work_dir)



from express.parsers.apps.espresso.settings import XML_DATA_FILE as ESPRESSO_XML_FILE
from express.parsers.apps.vasp.settings import XML_DATA_FILE as VASP_XML_FILE

from exaparser.utils import find_file, find_file_with_pattern
from .shell import ShellExecutionUnit, ESPRESSO_INPUT_FILE_REGEX
from ..modeling import ModelingExecutionUnit


class ShellWithResultsExecutionUnit(ShellExecutionUnit, ModelingExecutionUnit):
    """
    Shell unit parser class.

    Args:
        config (dict): unit config.
        work_dir (str): full path to working directory.
    """

    def __init__(self, config, work_dir):
        super(ShellWithResultsExecutionUnit, self).__init__(config, work_dir)

    def _is_vasp_calculation(self):
        if find_file("INCAR", self.work_dir):
            return True
        if find_file("POSCAR", self.work_dir):
            return True
        if find_file(VASP_XML_FILE, self.work_dir):
            return True

    def _is_espresso_calculation(self):
        if find_file(ESPRESSO_XML_FILE, self.work_dir):
            return True
        if find_file_with_pattern(ESPRESSO_INPUT_FILE_REGEX, self.work_dir):
            return True

    @property
    def parser_name(self):
        """
        Returns the name of the parser to pass to ExPrESS.

        Returns:
             str: espresso or vasp
        """
        if self._is_vasp_calculation():
            return "vasp"
        if self._is_espresso_calculation():
            return "espresso"



import os

from exaparser.job import Job
from exaparser.utils import read_json
from tests.enums import FIXTURES_DIR
from tests.integration import IntegrationTestBase


class TestJobParser(IntegrationTestBase):

    def setUp(self):
        super(TestJobParser, self).setUp()

    def tearDown(self):
        super(TestJobParser, self).setUp()

    def _clean_job_config(self, config):
        # These non-deterministic values in the results need to be overwritten with mock values, so
        # that our test comparisons work.
        config["workDir"] = "/mock/path"
        config["workflow"]["subworkflows"][0]["units"][0]["statusTrack"][0]["trackedAt"] = 1234567890

    def test_espresso_001_shell_job(self):
        """
        Extracts a job from an espresso calculation and asserts the results.
        """
        expected = read_json(os.path.join(FIXTURES_DIR, "espresso", "shell-job.json"))

        actual = Job("External Job", os.path.join(FIXTURES_DIR, "espresso", "test-001")).to_json()
        self._clean_job_config(actual)

        self.assertDeepAlmostEqual(expected, actual)

    def test_vasp_001_shell_job(self):
        """
        Extracts a job from a vasp calculation and asserts the results.
        """
        expected = read_json(os.path.join(FIXTURES_DIR, "vasp", "shell-job.json"))

        actual = Job("External Job", os.path.join(FIXTURES_DIR, "vasp", "test-001")).to_json()
        self._clean_job_config(actual)

        self.assertDeepAlmostEqual(expected, actual)



from express import ExPrESS

from .. import BaseExecutionUnit


class ModelingExecutionUnit(BaseExecutionUnit):
    def __init__(self, config, work_dir):
        super(ModelingExecutionUnit, self).__init__(config, work_dir)

    @property
    def parser_name(self):
        """
        Returns the name of the parser to pass to ExPrESS.

        Returns:
             str
        """
        raise NotImplementedError

    def safely_extract_property(self, property_, *args, **kwargs):
        """
        Safely extracts property.

        Args:
            property_ (str): property name.
            args (list): args passed to property extractor.
            kwargs (dict): kwargs passed to property extractor.

        Returns:
             dict
        """
        try:
            express = ExPrESS(self.parser_name, **dict(work_dir=self.work_dir, stdout_file=self.stdout_file))
            return express.property(property_, *args, **kwargs)
        except:
            pass

    @property
    def initial_structures(self):
        """
        Safely returns a list of initial structures used in this unit.

        Returns:
             list
        """
        structures = []
        try:
            express = ExPrESS(self.parser_name, **dict(work_dir=self.work_dir, stdout_file=self.stdout_file))
            for structure_string in express.parser.initial_structure_strings():
                initial_structure = express.property("material", structure_string=structure_string)
                initial_structure["name"] = "initial_structure"
                structures.append(initial_structure)
        except:
            pass
        return structures

    @property
    def final_structures(self):
        """
        Safely returns a list of final structures generated in this unit.

        Returns:
             list
        """
        structures = []
        try:
            express = ExPrESS(self.parser_name, **dict(work_dir=self.work_dir, stdout_file=self.stdout_file))
            for structure_string in express.parser.final_structure_strings():
                final_structure = express.property("material", structure_string=structure_string)
                final_structure["name"] = "final_structure"
                final_structure["repetition"] = 0
                structures.append(final_structure)
        except:
            pass
        return structures

    @property
    def structures(self):
        """
        Returns a list of structure pairs (initial/final) extracted from the unit.

        Returns:
             list[dict]
        """
        structures = []
        final_structures = self.final_structures
        initial_structures = self.initial_structures
        for index, structure in enumerate(initial_structures):
            structures.append({
                "initial": structure,
                "final": final_structures[index]
            })
        return structures

    @property
    def properties(self):
        """
        Returns a list of properties in EDC format extracted from the unit.

        Returns:
             list[dict]
        """
        properties = []
        for name in [r["name"] for r in self.results]:
            if name == "final_structure":
                continue
            property_ = self.safely_extract_property(name)
            if property_:
                property_.update({"repetition": 0})
                properties.append({
                    "data": property_,
                    "source": {
                        "type": "exabyte",
                        "info": {
                            "unitId": self.flowchartId
                        }
                    }
                })
        return properties



import re
import xml.etree.ElementTree as ET

from express.parsers.apps.espresso.settings import XML_DATA_FILE as ESPRESSO_XML_FILE

from exaparser.enums import *
from exaparser.utils import find_file
from . import ModelingExecutionUnit


class EspressoExecutionUnit(ModelingExecutionUnit):
    """
    Espresso execution unit parser class.

    Args:
        config (dict): unit config.
        work_dir (str): full path to working directory.
    """

    def __init__(self, config, work_dir):
        super(EspressoExecutionUnit, self).__init__(config, work_dir)
        self.xml_path = find_file(ESPRESSO_XML_FILE, self.work_dir)

    @property
    def parser_name(self):
        """
        Returns the name of the parser to pass to ExPrESS.

        Returns:
             str
        """
        return "espresso"

    @property
    def version(self):
        """
        Returns the application version used in the unit.

        Returns:
             str
        """
        root = ET.parse(self.xml_path).getroot()
        version = root.find('HEADER').find("CREATOR").attrib.get("VERSION").strip()
        return version if version in ESPRESSO_SUPPORTED_VERSIONS else ESPRESSO_DEFAULT_VERSION

    @property
    def application(self):
        """
        Returns the application used in the unit.

        Returns:
             dict
        """
        return {
            "name": "espresso",
            "version": self.version,
            "summary": "Quantum Espresso"
        }

    @property
    def executable(self):
        """
        Returns the executable used in the unit.

        Returns:
             dict
        """
        with open(self.stdout_file) as f:
            executable = ESPRESSO_EXECUTABLE_NAME_MAP[re.findall(ESPRESSO_EXECUTABLE_NAME_REGEX, f.read())[0]]
            return {
                "name": executable
            }



from . import BaseUnit


class SubworkflowUnit(BaseUnit):
    """
    Subworkflow unit parser class.

    Args:
        config (dict): unit config.
        work_dir (str): full path to working directory.
    """

    def __init__(self, config, work_dir):
        super(SubworkflowUnit, self).__init__(config, work_dir)
        self.id = self.config["_id"]

    def to_json(self):
        """
        Returns the subworkflow unit in JSON format.

        Returns:
             dict
        """
        config = super(SubworkflowUnit, self).to_json()
        config.update({
            "_id": self.id
        })
        return config



import json
import os
import re
from multiprocessing import Pool

from exabyte_api_client.endpoints.jobs import JobEndpoints
from exabyte_api_client.endpoints.materials import MaterialEndpoints
from exabyte_api_client.endpoints.projects import ProjectEndpoints
from exabyte_api_client.endpoints.raw_properties import RawPropertiesEndpoints

from exaparser.config import ExaParserConfig
from exaparser.enums import OUTPUT_CHUNK_SIZE
from exaparser.utils import upload_file_to_object_storage
from .. import DataHandler


class ExabyteRESTFulAPIDataHandler(DataHandler):
    """
    Exabyte RESTFul API data handler class.


    Args:
        job (exaparser.job.Job)
    """

    def __init__(self, job):
        super(ExabyteRESTFulAPIDataHandler, self).__init__(job)
        self._owner = self._project = None
        self.job_endpoints = JobEndpoints(*self.endpoint_args)
        self.project_endpoints = ProjectEndpoints(*self.endpoint_args)
        self.material_endpoints = MaterialEndpoints(*self.endpoint_args)
        self.raw_properties_endpoints = RawPropertiesEndpoints(*self.endpoint_args)

    @property
    def endpoint_args(self):
        """
        Returns a list of arguments passed to the endpoints.

        Returns:
             list
        """
        return [
            ExaParserConfig.get("exabyte_api_data_handler", "hostname"),
            ExaParserConfig.getint("exabyte_api_data_handler", "port"),
            ExaParserConfig.get("exabyte_api_data_handler", "account_id"),
            ExaParserConfig.get("exabyte_api_data_handler", "auth_token"),
            ExaParserConfig.get("exabyte_api_data_handler", "version"),
            ExaParserConfig.getboolean("exabyte_api_data_handler", "secure")
        ]

    @property
    def owner(self):
        """
        Returns the owner configuration extracted from the project.

        Returns:
             dict
        """
        if not self._owner:
            self._owner = self.project["owner"]
        return self._owner

    @property
    def project(self):
        """
        Returns the project the job should be created in.

        Returns:
             dict
        """
        if not self._project:
            self._project = self.project_endpoints.list(
                {
                    "slug": ExaParserConfig["global"]["project_slug"],
                    "owner.slug": ExaParserConfig["global"]["owner_slug"]
                }
            )[0]
        return self._project

    def create_structures(self, job):
        """
        Creates initial/final structures in web application.

        Returns:
             list
        """
        structures = []
        for config in self.job.structures:
            config["final"]["owner"] = self.owner
            config["final"]["tags"] = ["external"]
            config["initial"]["owner"] = self.owner
            config["initial"]["tags"] = ["external"]
            structures.append(config)

        if not len(structures):
            return []
        headers = self.material_endpoints.headers
        data = json.dumps({
            "jobId": job["_id"],
            "structures": structures,
            "unitId": job["workflow"]["subworkflows"][0]["units"][0]["flowchartId"],
        })

        return self.material_endpoints.request('POST', "structures/", headers=headers, data=data)

    def create_job(self):
        """
        Creates job in web application.

        Note: job must be set as external (isExternal=True) to be able to access the files.

        Returns:
             dict
        """
        config = self.job.to_json()
        config["isExternal"] = True
        config["_materials"] = []
        config["owner"] = self.owner
        config["_project"] = {"_id": self.project["_id"]}
        return self.job_endpoints.create(config)

    def update_job(self, job, structures):
        """
        Updates the job with given structures.

        Returns:
             dict
        """
        job["_materials"] = [{"_id": _id} for _id in list(set([c["initial"]["_id"] for c in structures]))]
        return self.job_endpoints.update(job["_id"], job)

    def create_properties(self, job_id):
        """
        Creates properties in web application. Properties will be shown in job's results tab after creation.

        Args:
            job_id (str): job ID.
        """
        for property_ in self.job.properties:
            property_["source"]["info"]["jobId"] = job_id
            self.raw_properties_endpoints.create(property_)

    @property
    def files(self):
        """
        Returns a list file paths relative to jb working directory to upload to object storage.

        Note: files matching excluded_files_regex are ignored.

        Returns:
             list
        """
        files_ = []
        for root, dirs, files in os.walk(self.job.work_dir):
            for file_ in [os.path.join(root, f) for f in files]:
                regex = ExaParserConfig.get("exabyte_api_data_handler", "excluded_files_regex")
                if regex and re.match(regex, file_):
                    continue
                files_.append(file_.replace("".join((self.job.work_dir, "/")), ""))
        return files_

    def upload_files(self, job_id):
        """
        Uploads the file in parallel into object storage.

        Note: permission to upload files into object storage is given through presignedURLs.

        Args:
            job_id (str): job ID.
        """
        presigned_urls = self.job_endpoints.get_presigned_urls(job_id, self.files)
        presigned_urls = [{"path": os.path.join(self.job.work_dir, p["file"]), "URL": p["URL"]} for p in presigned_urls]
        num_workers = min(len(presigned_urls), ExaParserConfig.getint("exabyte_api_data_handler", "num_workers"))
        pool = Pool(processes=num_workers)
        pool.map(upload_file_to_object_storage, presigned_urls)

    def upload_stdout(self, job_id):
        """
        Uploads stdout file.

        Args:
            job_id (str): job ID.
        """
        for config in self.job.stdout_files:
            output_chunks_count = 0
            file_descriptor = open(config["stdoutFile"])
            while True:
                chunk = file_descriptor.read(OUTPUT_CHUNK_SIZE)
                if chunk != '':
                    data = {
                        "chunk": chunk,
                        "repetition": 0,
                        "order": output_chunks_count,
                        "unitFlowchartId": config["unitFlowchartId"],
                    }
                    self.job_endpoints.insert_output_files(id_=job_id, data=json.dumps(data))
                    output_chunks_count += 1
                else:
                    break

    def handle(self):
        """
        Creates materials, job and properties in webapp and uploads files into object storage.
        """
        job = self.create_job()
        structures = self.create_structures(job)
        if len(structures):
            self.update_job(job, structures)
            self.create_properties(job["_id"])
        if ExaParserConfig.getboolean("exabyte_api_data_handler", "upload_files"):
            self.upload_files(job["_id"])
        if ExaParserConfig.getboolean("exabyte_api_data_handler", "upload_stdout_file"):
            self.upload_stdout(job["_id"])



from .. import TestBase


class UnitTestBase(TestBase):
    """
    Base class for ExParser unit tests.
    """

    def setUp(self):
        super(UnitTestBase, self).setUp()

    def tearDown(self):
        super(UnitTestBase, self).setUp()



class BaseUnit(object):
    """
    Base unit parser class.

    Args:
        config (dict): unit config.
        work_dir (str): full path to working directory.
    """

    def __init__(self, config, work_dir):
        self.config = config
        self.work_dir = work_dir
        self.type = self.config["type"]
        self.name = self.config.get("name", "")
        self.head = self.config.get("head", False)
        self.flowchartId = self.config["flowchartId"]
        self.next_flowchartId = self.config.get("next", "")

    def to_json(self):
        """
        Returns the unit in JSON format.

        Returns:
             dict
        """
        return {
            "flowchartId": self.flowchartId,
            "name": self.name,
            "head": self.head,
            "type": self.type,
            "next": self.next_flowchartId
        }



from .. import BaseComputeParser


class PBSComputeParser(BaseComputeParser):
    """
    PBS/Torque compute parser class.

    Args:
        work_dir (str): full path to working directory.
    """

    def __init__(self, work_dir):
        super(PBSComputeParser, self).__init__(work_dir)



class BaseComputeParser(object):
    """
    Base compute parser class.

    Args:
        work_dir (str): full path to working directory.
    """

    def __init__(self, work_dir):
        self.work_dir = work_dir

    @property
    def ppn(self):
        """
        Returns number of processors per node.
        Override on child classes.

        Returns:
             int
        """
        return 1

    @property
    def nodes(self):
        """
        Returns number of nodes.
        Override on child classes.

        Returns:
             int
        """
        return 1

    @property
    def queue(self):
        """
        Returns queue name.
        Override on child classes.

        Returns:
             int
        """
        return "D"

    @property
    def walltime(self):
        """
        Returns job walltime.
        Override on child classes.

        Returns:
             int
        """
        return "01:00:00"

    def to_json(self):
        return {
            "ppn": self.ppn,
            "nodes": self.nodes,
            "queue": self.queue,
            "timeLimit": self.walltime
        }



from .subworkflow import Subworkflow
from .units.factory import get_unit


class Workflow(object):
    """
    Workflow parser class.

    Args:
        config (dict): workflow config.
        work_dir (str): full path to working directory.
    """

    def __init__(self, config, work_dir):
        self.config = config
        self.work_dir = work_dir
        self.units = [get_unit(config, self.work_dir) for config in self.config.get("units", [])]
        self.subworkflows = [Subworkflow(config, self.work_dir) for config in self.config.get("subworkflows", [])]

    @property
    def name(self):
        """
        Returns workflow name.

        Returns:
             str
        """
        return self.config.get("name", "Workflow")

    @property
    def execution_units(self):
        """
        Returns a list of all execution units in this workflow.

        Returns:
             list
        """
        execution_units = []
        for subworkflow in self.subworkflows:
            execution_units.extend(subworkflow.execution_units)
        return execution_units

    def to_json(self):
        """
        Returns the workflow in JSON format.

        Returns:
             dict
        """
        return {
            "name": self.name,
            "properties": [],
            "units": [u.to_json() for u in self.units],
            "subworkflows": [u.to_json() for u in self.subworkflows],
            "isMultiMaterial": True  # to let job have multiple materials
        }



from exaparser.enums import *
from . import ScriptingExecutionUnit


class ShellExecutionUnit(ScriptingExecutionUnit):
    """
    Shell unit parser class.

    Args:
        config (dict): unit config.
        work_dir (str): full path to working directory.
    """

    def __init__(self, config, work_dir):
        super(ShellExecutionUnit, self).__init__(config, work_dir)

    @property
    def application(self):
        """
        Returns the application used in the unit.

        Returns:
             dict
        """
        return {
            "name": "shell",
            "summary": "Shell Script",
            "version": self.version
        }

    @property
    def version(self):
        """
        Returns the application version used in the unit.

        Returns:
             str
        """
        return SHELL_DEFAULT_VERSION

    @property
    def executable(self):
        """
        Returns the executable used in the unit.

        Returns:
             dict
        """
        return {
            "name": "sh"
        }



from .managers.pbs import PBSComputeParser
from .managers.slurm import SLURMComputeParser


def get_compute_parser(name, work_dir):
    """
    Returns an instance of compute parser class

    Args:
        name (str): parser name, PBS or SLURM.
        work_dir (str): full path to job working directory.
    """
    parsers = dict(PBS=PBSComputeParser, SLURM=SLURMComputeParser)
    return parsers[name](work_dir)



import argparse

from .config import ExaParserConfig
from .data.factory import get_data_handler
from .job import Job


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument('-n', '--name', required=True, help='job name')
    parser.add_argument('-c', '--config', help='full path to config file')
    parser.add_argument('-w', '--work-dir', dest="work_dir", required=True, help='full path to working directory')
    return parser.parse_args()


def main():
    args = parse_arguments()
    if args.config:
        ExaParserConfig.read(args.config)
    job = Job(args.name, args.work_dir)
    for handler in ExaParserConfig.get("global", "data_handlers").replace(" ", "").split(","):
        get_data_handler(handler, job).handle()
